# Copyright 2014 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import os

import ckx_tools.metadata as metadata

from ckx_tools.argument_parsing import add_cmake_and_make_and_catkin_make_args
from ckx_tools.argument_parsing import add_context_args

from ckx_tools.context import Context

from ckx_tools.terminal_color import ColorMapper
from audioop import cross

from . import utilities

color_mapper = ColorMapper()
clr = color_mapper.clr


def prepare_arguments(parser):

    parser.description = "This verb is used to configure a catkin workspace's\
    configuration and layout. Calling `catkin config` with no arguments will\
    display the current config and affect no changes if a config already exists\
    for the current workspace and profile."

    # Workspace / profile args
    add_context_args(parser)

    common_group = parser.add_argument_group('Common Options', 'Most frequently used options.')
    # add = common_group.add_argument
    add = common_group.add_mutually_exclusive_group().add_argument
    add('--underlays', '-u', dest='underlays', type=str, default=None,
        help='Explicitly extend the result-space of other catkin workspace underlays, '
        'overriding the value of $CMAKE_PREFIX_PATH.')
    add('--no-underlays', dest='underlays', action='store_const', const='',
        help="Clear any setting previously created by '--underlays'")
    add('-i', '--install-space', default=None,
        help="Sets the target install space ['{0}']".format(Context.DEFAULT_INSTALL_SPACE))
    add('--install', action='store_true', default=None,
        help='Enables the installation step')
    add('--no-install', dest='install', action='store_false', default=None,
        help='Disables the installation step')
    add = common_group.add_mutually_exclusive_group().add_argument
    add('--whitelist', metavar="PKG", dest='whitelist', nargs="+", required=False, type=str, default=None,
        help='Set the packages on the whitelist. If the whitelist is non-empty, '
        'only the packages on the whitelist are built with a bare call to '
        '`catkin build`.')
    add('--no-whitelist', dest='whitelist', action='store_const', const=[], default=None,
        help='Clear all packages from the whitelist.')
    add = common_group.add_mutually_exclusive_group().add_argument
    add('--blacklist', metavar="PKG", dest='blacklist', nargs="+", required=False, type=str, default=None,
        help='Set the packages on the blacklist. Packages on the blacklist are '
        'not built with a bare call to `catkin build`.')
    add('--no-blacklist', dest='blacklist', action='store_const', const=[], default=None,
        help='Clear all packages from the blacklist.')
    add('--init', action='store_true', default=False,
        help='Initialize a workspace if it does not yet exist.')

    spaces_group = parser.add_argument_group('Advanced Space Configuration', 'Location of parts of the catkin workspace.')
    add = spaces_group.add_argument
    add('-s', '--source-space', default=Context.DEFAULT_SOURCE_SPACE,
        help="The path to the source space ['{0}']".format(Context.DEFAULT_SOURCE_SPACE))
    add('-l', '--log-space', default=None,
        help="The path to the log space ['{0}']".format(Context.DEFAULT_LOG_SPACE))
    add('-b', '--build-space', default=None,
        help="The path to the build space ['{0}']".format(Context.DEFAULT_BUILD_SPACE))
    add('-d', '--devel-space', default=None,
        help="Sets the target devel space ['{0}']".format(Context.DEFAULT_DEVEL_SPACE))

    devel_group = parser.add_argument_group(
        'Advanced Devel Configuration', 'Options for configuring the structure of the devel space.')
    add = devel_group.add_mutually_exclusive_group().add_argument
    add('--link-devel', dest='devel_layout', action='store_const', const='linked', default=None,
        help='Build products from each catkin package into isolated spaces,'
        ' then symbolically link them into a merged devel space.')
    add('--merge-devel', dest='devel_layout', action='store_const', const='merged', default=None,
        help='Build products from each catkin package into a single merged devel spaces.')
    add('--isolate-devel', dest='devel_layout', action='store_const', const='isolated', default=None,
        help='Build products from each catkin package into isolated devel spaces.')

    install_group = parser.add_argument_group(
        'Advanced Install Configuration', 'Options for configuring the structure of the install space.')
    add = install_group.add_mutually_exclusive_group().add_argument
    add('--isolate-install', action='store_true', default=None,
        help='Each package to a separate install space.')
    add('--merge-install', dest='isolate_install', action='store_false', default=None,
        help='Each package to a single merged install space [default].')

    build_group = parser.add_argument_group('Advanced Build Options', 'Options for configuring the way packages are built.')
    add_cmake_and_make_and_catkin_make_args(build_group)

    cross_compiling_group = parser.add_argument_group('Cross Compiling', 'Options for configuring a cross compiling environment.')
    add = cross_compiling_group.add_argument
    add('--list-toolchains', action='store_true', help='list all currently available toolchain modules [false]')
    add('--list-platforms', action='store_true', help='list all currently available platform modules [false]')
    add('--toolchain', action='store', default=None, help='toolchain cmake module to load []')
    add('--platform', action='store', default=None, help='platform cmake cache module to load [default]')

    behavior_group = parser.add_argument_group('Argument handling', 'Options affecting argument handling.')
    add = behavior_group.add_mutually_exclusive_group().add_argument
    add('--append-args', '-a', action='store_true', default=False,
        help='For list-type arguments, append elements.')
    add('--remove-args', '-r', action='store_true', default=False,
        help='For list-type arguments, remove elements.')

    return parser

def main(opts):
    try:
        # Determine if the user is trying to perform some action, in which
        # case, the workspace should be automatically initialized
        ignored_opts = ['main', 'verb']
        actions = [v for k, v in vars(opts).items() if k not in ignored_opts]
        no_action = not any(actions)

        # Generic display functions only
        # List toolchains or platforms
        if opts.list_toolchains:
            utilities.list_toolchains()
            return 0
        if opts.list_platforms:
            utilities.list_platforms()
            return 0

        # Try to find a metadata directory to get context defaults
        # Otherwise use the specified directory
        context = Context.load(
            opts.workspace,
            opts.profile,
            opts,
            append=opts.append_args,
            remove=opts.remove_args)

        do_init = opts.init or not no_action
        summary_notes = []

        if not context.initialized() and do_init:
            summary_notes.append(clr('@!@{cf}Initialized new catkin workspace in `%s`@|' % context.workspace))

        if context.initialized() or do_init:
            Context.save(context)
            try:
                config_doc_prefix = ""
                utilities.instantiate_or_update_config_environment(
                    context.profile,
                    context.workspace,
                    context.build_root_abs,
                    context.platform,
                    context.toolchain,
                    config_doc_prefix,
                    context.underlays
                )
            except RuntimeError as e:
                print(clr("[config] @!@{rf}Error:@| %s") % e.message)
                return 1

        if not context.source_space_exists():
            os.makedirs(context.source_space_abs)

        print(context.summary(notes=summary_notes))

    except IOError as e:
        # Usually happens if workspace is already underneath another ckx_tools workspace
        print(clr("[config] @!@{rf}Error:@| could not configure catkin workspace: `%s`") % e.message)
        return 1

    return 0



#!/usr/bin/env python
#
# License: BSD
#   https://raw.github.com/stonier/ckx_tools/devel/LICENSE
#
##############################################################################
# Documentation
##############################################################################

"""
Implementation of the 'ckx rosdep' verb.
"""

##############################################################################
# Imports
##############################################################################

import os
import subprocess
import sys
import traceback

try:
    from queue import Queue  # Python3
except ImportError:
    from Queue import Queue  # Python2

import ckx_tools.argument_parsing as argument_parsing
import ckx_tools.common as common
import ckx_tools.execution.job_server as job_server
import ckx_tools.metadata as metadata

from ckx_tools.context import Context
from ckx_tools.execution.executor import execute_jobs
from ckx_tools.execution.executor import run_until_complete
from ckx_tools.execution.jobs import Job
from ckx_tools.execution.stages import CommandStage
from ckx_tools.execution.controllers import ConsoleStatusController
from ckx_tools.terminal_color import ColorMapper
from ckx_tools.utils import which

color_mapper = ColorMapper()
clr = color_mapper.clr

##############################################################################
# Entry Point API
##############################################################################


def help_string():
    instructions = clr("@!Examples@|\n\n  \
@{gf}Minimal Use Cases@|\n\n  \
  @{cf}ckx rosdep --install@| : @{yf}install rosdeps for the active profile in the enclosing workspace@|\n  \
  @{cf}ckx rosdep --list@|    : @{yf}list all rosdep keys for the active profile in the enclosing workspace@|\n\n  \
@{gf}Targeted Profile or Workspace@|\n\n  \
  @{cf}ckx rosdep --workspace ~/foo_ws --install@| : @{yf}install rosdeps for the active profile in the specified workspace@|\n  \
  @{cf}ckx rosdep --profile native --install@| : @{yf}install rosdeps for the specified profile in the enclosing workspace@|\n  \
 ")
    return instructions


def prepare_arguments(parser):
    default_track = common.get_default_track()

    parser.epilog = help_string()
    argument_parsing.add_context_args(parser)  # workspace / profile args
    add = parser.add_argument

    add('--install', '-i', action='store_true', default=False,
        help='Install all rosdeps for this build profile (including underlays)')
    add('--keys', '-k', action='store_true', default=False,
        help='List all rosdep keys required by this build profile (including underlays)')
    add('--track', action='store', default=default_track, help='force the track (rosdistro) if the default guess is not accurate %s [%s]' % (common.VALID_TRACKS, default_track))
    return parser


def main(opts):
    # Are we in a parallel build profile? If so, prefer that over the active one
    if not opts.profile:
        enclosing_profile = metadata.find_enclosing_profile(os.getcwdu(), opts.workspace)
        if enclosing_profile:
            print(clr("@!\nInfo: in a parallel build folder, prefer this profile [%s]\n@|" % enclosing_profile))
            opts.profile = enclosing_profile

    # Load the context
    context = Context.load(opts.workspace, opts.profile, opts)

    # Initialize job server
    job_server.initialize(
        max_jobs=1,
        max_load=None,
        gnu_make_enabled=False)

    underlays = context.cmake_prefix_path.split(';')
    if opts.install:
        # alternative: the ckx tools job server way, but rosdep install needs sudo and this adds root logs!
        # create_and_execute_job('install', cmd, context.workspace, context.log_space_abs)
        # NOTE: don't use log here - user is probably running sudo, prefer sys.stdout.write
        sys.stdout.write(clr("@!Installing Rosdeps@|\n"))
        cmd = rosdeps_install_command(context, underlays, opts.track)
        try:
            proc = subprocess.Popen(
                cmd, cwd=context.source_space_abs, shell=False,
                stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=os.environ.copy()
            )
        except OSError as e:
            sys.stdout.write(clr("@{rf}[ERROR] failed command '{0}': {1}").format(cmd, e))
            return 1
        error_flag = False
        while True:
            line = proc.stdout.readline().decode('utf8', 'replace')
            line = unicode(line)
            if proc.returncode is not None or not line:
                break
            if line.startswith("ERROR"):
                error_flag = True
            if error_flag:
                sys.stdout.write(clr("@{rf}" + line + "@!"))
            else:
                sys.stdout.write(line)
        proc.wait()
        # perhaps check proc.returncode and print ERROR if true
        return 0
    if opts.keys:
        sys.stdout.write(clr("@!List Rosdep Keys@|\n"))
        cmd = rosdeps_keys_command(context, underlays, opts.track)
        try:
            proc = subprocess.Popen(
                cmd, cwd=context.source_space_abs, shell=False,
                stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=os.environ.copy()
            )
        except OSError as e:
            sys.stdout.write(clr("@{rf}[ERROR] failed command '{0}': {1}").format(cmd, e))
            return 1
        proc.wait()
        out, err = proc.communicate()
        if err:
            sys.stdout.write(clr("@{rf}" + err + "@!"))
        else:
            sys.stdout.write(clr("@{gf}  Keys@!\n"))
            rosdep_keys = [s for s in out.split("\n") if s]  # drop empty lines
            for key in sorted(rosdep_keys):
                sys.stdout.write(clr("   @{cf}+@{yf} {0}\n").format(key))

        # create_and_execute_job('keys', cmd, context.workspace, context.log_space_abs)
        return 0
    return 0

##############################################################################
# Helpers
##############################################################################


def extend_cmd_from_paths(underlays):
    extended_cmd = []
    sys.stdout.write(clr("  @{gf}From Paths@!\n"))
    for underlay in underlays:
        underlay_path = underlay
        # if it is a devel workspace -> hunt for the sources
        enclosing_workspace = metadata.find_enclosing_workspace(underlay_path)
        if enclosing_workspace:
            underlay_source_path = os.path.abspath(os.path.join(enclosing_workspace, "src"))
            if os.path.isdir(underlay_source_path):
                underlay_path = underlay_source_path
        if os.path.isdir(underlay_path):
            extended_cmd += ['--from-paths', underlay_path]
            sys.stdout.write(clr("   @{cf}+@{yf} {0}\n").format(underlay_path))
        else:
            sys.stdout.write(clr("   @{cf}-@{rf} {0} [not found]\n").format(underlay_path))
    return extended_cmd


def rosdeps_keys_command(context, underlays, rosdistro):
    """
    :param str underlays: comma separated string of underlays
    :returns: list of strings comprising the command to execute
    """
    source_path = os.path.abspath(os.path.join(context.workspace, "src"))
    cmd = [which('rosdep'), 'keys', '-r']
    cmd += extend_cmd_from_paths(underlays)
    cmd += ['--from-paths', source_path, '--ignore-src', '--rosdistro', rosdistro]
    return cmd


def rosdeps_install_command(context, underlays, rosdistro):
    """
    :param str underlays: comma separated string of underlays
    :returns: list of strings comprising the command to execute
    """
    source_path = os.path.abspath(os.path.join(context.workspace, "src"))

    cmd = [which('rosdep'), 'install']
    cmd += extend_cmd_from_paths(underlays)
    cmd += ['--from-paths', source_path, '--ignore-src', '--rosdistro', rosdistro, '-y']
    return cmd


def create_and_execute_job(job_name, cmd, workspace_path, log_path):
    """
    Put the rosdep job on the job server.
    :param str job_name: user friendly label for the job
    :param [str] cmd: actual command to execute
    """
    stages = []
    stages.append(CommandStage(
        job_name,
        cmd,
        cwd=workspace_path,
        occupy_job=True
    ))
    job = Job(
        jid="rosdep",  # unique job identifier
        deps=[],
        env_loader=load_env,  # get_env_loader(package, context),
        stages=stages)
    event_queue = Queue()
    status_thread = ConsoleStatusController(
        'rosdep',
        job_labels=['rosdep', 'install'],
        jobs=[job],
        max_toplevel_jobs=1,
        available_jobs=[],  # [pkg.name for _, pkg in context.packages],
        whitelisted_jobs=[],
        blacklisted_jobs=[],
        event_queue=event_queue,
        show_buffered_stdout=True,
    )
    status_thread.start()
    # Block while running N jobs asynchronously
    try:
        unused_all_succeeded = run_until_complete(execute_jobs(
            'rosdep',
            jobs=[job],
            locks={},
            event_queue=event_queue,
            log_path=log_path,
            max_toplevel_jobs=1
        ))
    except Exception:
        status_thread.keep_running = False
        status_thread.join(1.0)
        common.wide_log(str(traceback.format_exc()))
    status_thread.join(1.0)

    return 0


def load_env(base_env):
    """
    Don't need an extended environment for this job. Just
    return the base environment directly.

    :param base_env: usually this is just os.environ as is
    """
    return base_env




#!/usr/bin/env python
#
# License: BSD
#   https://raw.github.com/stonier/ckx_tools/devel/LICENSE
#
##############################################################################
# Documentation
##############################################################################

"""
Implementation of the 'ckx ws' verb.
"""

##############################################################################
# Imports
##############################################################################

import argparse
import ckx_tools.metadata as metadata
import os
import sys
import urllib2
import urlparse
import vci
import vcstool.commands.import_
import vcstool.executor
import yaml

from ckx_tools.context import Context
from ckx_tools.terminal_color import ColorMapper
color_mapper = ColorMapper()
clr = color_mapper.clr

##############################################################################
# Entry Point API
##############################################################################


def help_string():
    overview = 'Creates and populates catkin workspaces.\n\n'
    instructions = "Usage:\n\n  \
Empty Workspace:\n  \
 - 'ckx ws ecl' : create an empty workspace in ./ecl.\n  \
 - 'ckx ws ~/ecl' : create an empty workspace in ~/ecl.\n  \
From Rosinstall:\n  \
 - 'ckx ws ecl ecl.repos' : populate a workspace from a .repos file.\n  \
 - 'ckx ws ecl https://raw.github.com/stonier/ecl_core/devel/ecl.repos' : populate from url.\n  \
From Rosinstall Database:\n  \
 - 'ckx ws ecl ecl' : populate a workspace from the currently set version control index.\n  \
 - 'ckx ws --merge ./ecl_extras.repos' : merge sources from another .repos installer.\n  \
Management:\n  \
 - 'ckx ws --clean' : clean metadata produced by 'ckx config', 'ckx build' from the workspace.\n  \
Configuration:\n  \
 - 'ckx ws --get-vci-url' : return the currently configured version control index url.\n  \
 - 'ckx ws --set-vci-url' : save this url as the default version control index url.\
 "
    return overview + instructions


def prepare_arguments(parser):
    parser.epilog = help_string()
    add = parser.add_argument
    add('dir', nargs='?', default=os.getcwd(), help='directory to use for the workspace [cwd]')
    add('-j', '--jobs', action='store', default=10, help='how many parallel threads to use for installing [10]')
    add('uri', nargs=argparse.REMAINDER, default=None, help='uri for a rosinstall file [None]')

    management_group = parser.add_argument_group('Management', 'options to assist in managing your workspace')
    add = management_group.add_argument
    add('--clean', action='store_true', default=False,
        help='Clean build metadata for the given workspace (does not touch rosinstalled sources).')

    configuration_group = parser.add_argument_group('Configuration', 'Options to assist in configuring this tool')
    add = configuration_group.add_argument
    add('--list', action='store_true', help='list all currently available repos in the index [false]')
    parser.add_argument('--get-vci-url', action='store_true', help='print the default rosinstall database uri')
    parser.add_argument('--set-vci-url', action='store', default=None, help='set a new default  rosinstall database uri')

    return parser


def main(opts):
    '''
      Process the workspace command and return success or failure to the calling script.
    '''

    ########################################
    # Version Control Index Management
    ########################################
    if opts.get_vci_url:
        print(clr("\n@{cf}URL @|: @{yf}{0}@|\n").format(vci.config.get_index_url()))
        return 0
    if opts.set_vci_url:
        vci.config.set_index_url(opts.set_vci_url)
        print(clr("\n@{cf}URL @|: @{yf}{0}@|\n").format(opts.set_vci_url))
        return 0
    if opts.list:
        url = vci.config.get_index_url()
        try:
            contents = vci.index_contents.get(url)
        except urllib2.URLError as e:
            print(clr("\n@{rf}[ERROR] could not retrieve {0}@|").format(str(e)))
            sys.exit(1)
        vci.index_contents.display(url, contents)
        return 0

    ########################################
    # Clean all build/profile information
    ########################################
    workspace_dir = opts.dir if os.path.isabs(opts.dir) else os.path.join(os.getcwd(), opts.dir)
    if opts.clean:
        profile_names = metadata.get_profile_names(workspace_dir)
        for name in profile_names:
            metadata.remove_profile(workspace_dir, name)
        metadata.init_metadata_root(
            workspace_path=workspace_dir,
            reset=True)
        return 0

    ########################################
    # Check/Init a Workspace
    ########################################
    initialised_new_workspace = False
    sources_dir = os.path.join(workspace_dir, 'src')
    try:
        # Try to load an existing context
        ctx = Context.load(workspace_dir, strict=True)

        # Initialize a new context if necessary
        if not ctx:
            print('Initializing catkin workspace in `%s`.' % (workspace_dir or os.getcwd()))
            if not os.path.isdir(workspace_dir):
                os.mkdir(workspace_dir)
            # Init the metadata
            metadata.init_metadata_root(
                workspace_path=workspace_dir or os.getcwd(),
                reset=False)
            ctx = Context.load(workspace_dir)
            os.mkdir(sources_dir)
            initialised_new_workspace = True
        else:
            print('[ws] catkin workspace `%s` found.' % (ctx.workspace))
        # print(ctx.summary())

    except IOError as exc:
        # Usually happens if workspace is already underneath another ckx_tools workspace
        print(clr("[ws] @!@{rf}Error:@| could not init/introspect the workspace (it's nested under another?): `%s`" % exc.message))
        return 1

    ####################
    # Rosinstalls
    ####################
    file_list = []
    uri_list = []
    lookup_name_list = []
    lookup_index_url = vci.config.get_index_url()
    if opts.uri:
        # parse the list looking for 1) abs 2) rel 3) lookup names, 4) http uri's
        for uri in opts.uri:
            if os.path.isabs(uri):
                file_list.append(uri)
            elif os.path.isfile(os.path.join(os.getcwd(), uri)):
                file_list.append(os.path.join(os.getcwd(), uri))
            elif urlparse.urlparse(uri).scheme == "":  # not a http element, let's look up our database
                lookup_name_list.append(uri)
            else:  # it's a http element'
                uri_list.append(uri)
#         # lookup the database and convert to http uri's if necessary
#         if lookup_name_list:
#             try:
#                 rosinstall_database, lookup_database = get_rosinstall_database(opts.track)
#             except RuntimeError as exc:
#                 print(clr("[ws] @!@{rf}Error: %s@|" % str(exc)))
#                 return 1
#             (database_name_list, database_uri_list) = parse_database(lookup_name_list, rosinstall_database)
#             lookup_name_list.extend(database_name_list)
#             uri_list.extend(database_uri_list)
        populate_sources(sources_dir, file_list, uri_list, lookup_name_list, opts.jobs)
        print_details(workspace_dir,
                      file_list,
                      uri_list,
                      lookup_name_list,
                      lookup_index_url,
                      initialised_new_workspace
                      )
    elif initialised_new_workspace:
        print_banner("Initialised Empty Workspace")
        print(clr("@{cf}Workspace  : @|@{yf}{0}@|").format(workspace_dir))
        print_footer()
        print_proceed()
    else:
        print_banner("Workspace Details")
        print(clr("@{cf}Workspace  : @|@{yf}{0}@|").format(workspace_dir))
        print_footer()

    return 0

##############################################################################
# Settings
##############################################################################

DEFAULT_ROSINSTALL_DATABASE = 'https://raw.github.com/stonier/ckx_tools/rosinstalls'


def _execute_vci(args):
    """
    :param [str] args: list of args to pass to vci
    """

##############################################################################
# Helpers
##############################################################################


class Args(dict):
    """
    Example:
    m = Args({'first_name': 'Eduardo'}, last_name='Pool', age=24, sports=['Soccer'])
    """
    def __init__(self, *args, **kwargs):
        super(Args, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.iteritems():
                    self[k] = v

        if kwargs:
            for k, v in kwargs.iteritems():
                self[k] = v

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super(Args, self).__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super(Args, self).__delitem__(key)
        del self.__dict__[key]


def populate_sources(base_path, file_list, uri_list, lookup_name_list, parallel_jobs):
    '''
      :param str base_path: location of the vcstool workspace
      :param str file_list: list of files to setup
      :param str uri_list: list of url's to setup
      :param str lookup_name_list: list of vci keys to lookup and setup
      :param int parallel_jobs: faster, faster
    '''
    combined_yaml_contents = {'repositories': {}}
    if uri_list:
        print("@{yf}[WARN] populating from uri's is not supported at this time@|")
    for filename in file_list:
        print("Populating from filename {0}".format(filename))
        with open(filename, 'r') as stream:
            try:
                new_yaml_contents = yaml.load(stream)
                combined_yaml_contents['repositories'].update(new_yaml_contents['repositories'])
            except yaml.YAMLError as e:
                print(clr("@{yf}[WARN] could not open '{0}', skipping [{1}]@|").format(filename, str(e)))
    if lookup_name_list:
        print("Populating from lookup names {0}".format(lookup_name_list))
        new_yaml_contents = vci.find.create_yaml_from_key(lookup_name_list)
        combined_yaml_contents['repositories'].update(new_yaml_contents['repositories'])

    # the vcstool way (alternative : subprocess it)
    args = Args({'path': base_path, 'debug': False, 'workers': parallel_jobs, 'repos': True})
    repos = vcstool.commands.import_.get_repos_in_vcstool_format(combined_yaml_contents['repositories'])
    jobs = vcstool.commands.import_.generate_jobs(repos, args)  # need to wire up a special args here
    results = vcstool.executor.execute_jobs(jobs, show_progress=True, number_of_workers=parallel_jobs, debug_jobs=False)
    vcstool.executor.output_results(results)

##############################################################################
# Printers
##############################################################################


def print_banner(title):
    width = 80
    print(clr("@!\n" + "*" * width + "@|"))
    print(clr("@!{text:^{width}}".format(text=title, width=width) + "@|"))
    print(clr("@!*" * width + "@|"))


def print_footer():
    width = 80
    print(clr("@!*" * width + "\n" + "@|"))


def print_details(workspace_dir, file_list, uri_list, lookup_name_list, lookup_index_url, initialised_new_workspace):
    if initialised_new_workspace:
        print_banner("Initialised Workspace")
    else:
        print_banner("Updated Workspace")
    print(clr("@{cf}Workspace  : @|@{yf}" + workspace_dir + "@|"))
    if lookup_name_list:
        print(clr("@{cf}VCI Url    : @|@{yf}" + lookup_index_url + "@|"))
        s = "@{cf}VCI Keys   : @|@{yf}"
        for lookup_name in lookup_name_list:
            s += "%s " % lookup_name
        s += "@|"
        print(clr(s))
    if file_list:
        print(clr("@{cf}Files      : @|@{yf}%s@|" % file_list[0]))
        for filename in file_list[1:]:
            print(clr("@{cf}           : @|@{yf}%s@|" % filename))
    if uri_list:
        print(clr("@{cf}Rosisntalls: @|@{yf}%s@|" % uri_list[0]))
        for uri in uri_list[1:]:
            print(clr("@{cf}           : @|@{yf}%s@|" % uri))
    print_footer()
    if initialised_new_workspace:
        print_proceed()


def print_proceed():
    print(clr("@{cf}Proceed to configure parallel builds with 'ckx config'.@|\n"))



# Copyright 2014 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import os
import shutil

from ckx_tools.common import mkdir_p
from ckx_tools.common import get_cached_recursive_build_depends_in_workspace

from ckx_tools.resultspace import get_resultspace_environment

from ckx_tools.execution.events import ExecutionEvent


def get_env_loaders(package, context):
    """Get a list of env loaders required to build this package."""

    sources = []
    # If installing to isolated folders or not installing, but devel spaces are not merged
    if (context.install and context.isolate_install) or (not context.install and context.isolate_devel):
        # Source each package's install or devel space
        space = context.install_space_abs if context.install else context.devel_space_abs
        # Get the recursive dependcies
        depends = get_cached_recursive_build_depends_in_workspace(package, context.packages)
        # For each dep add a line to source its setup file
        for dep_pth, dep in depends:
            source_path = os.path.join(space, dep.name, 'env.sh')
            sources.append(source_path)
    else:
        # Get the actual destination of this package
        if context.link_devel and not context.install:
            source_path = os.path.join(context.package_final_path(package), 'env.sh')
        else:
            source_path = os.path.join(context.package_dest_path(package), 'env.sh')
        sources = [source_path]

    return sources


def get_env_loader(package, context):
    """This function returns a function object which extends a base environment
    based on a set of environments to load."""

    def load_env(base_env):
        # Copy the base environment to extend
        job_env = dict(base_env)
        # Get the paths to the env loaders
        env_loader_paths = get_env_loaders(package, context)
        # If DESTDIR is set, set _CATKIN_SETUP_DIR as well
        if context.destdir is not None:
            job_env['_CATKIN_SETUP_DIR'] = context.package_dest_path(package)

        for env_loader_path in env_loader_paths:
            # print(' - Loading resultspace env from: {}'.format(env_loader_path))
            resultspace_env = get_resultspace_environment(
                os.path.split(env_loader_path)[0],
                base_env=job_env,
                quiet=True,
                cached=context.use_env_cache,
                strict=False)
            job_env.update(resultspace_env)

        return job_env

    return load_env


def makedirs(logger, event_queue, path):
    """FunctionStage functor that makes a path of directories."""
    mkdir_p(path)
    return 0


def copyfiles(logger, event_queue, source_paths, dest_path):
    """FunctionStage functor that copies one or more files"""
    for source_path in source_paths:
        shutil.copy(source_path, dest_path)
    return 0


def rmfile(logger, event_queue, path):
    """FunctionStage functor that removes a file."""
    if os.path.exists(path):
        os.remove(path)
    return 0


def rmdirs(logger, event_queue, paths):
    """FunctionStage functor that removes a directory tree."""
    return rmfiles(logger, event_queue, paths, remove_empty=False)


def rmfiles(logger, event_queue, paths, dry_run, remove_empty=False, empty_root='/'):
    """FunctionStage functor that removes a list of files and directories.

    If remove_empty is True, then this will also remove directories which
    become emprt after deleting the files in `paths`. It will delete files up
    to the path specified by `empty_root`.
    """

    # Determine empty directories
    if remove_empty:
        # First get a list of directories to check
        dirs_to_check = set()

        for path in paths:
            # Make sure the file is given by an absolute path and it exists
            if not os.path.isabs(path) or not os.path.exists(path):
                continue

            # Only look in the devel space
            while empty_root.find(path) != 0:
                # Pop up a directory
                path, dirname = os.path.split(path)

                # Skip if this path isn't a directory
                if not os.path.isdir(path):
                    continue

                dirs_to_check.add(path)

        # For each directory which may be empty after cleaning, visit them
        # depth-first and count their descendants
        dir_descendants = dict()
        for path in sorted(dirs_to_check, key=lambda k: -len(k.split(os.path.sep))):
            # Get the absolute path to all the files currently in this directory
            files = [os.path.join(path, f) for f in os.listdir(path)]
            # Filter out the files which we intend to remove
            files = [f for f in files if f not in paths]
            # Compute the minimum number of files potentially contained in this path
            dir_descendants[path] = sum([
                (dir_descendants.get(f, 1) if os.path.isdir(f) else 1)
                for f in files
            ])

            # Schedule the directory for removal if removal of the given files will make it empty
            if dir_descendants[path] == 0:
                paths.append(path)

    # REmove the paths
    for index, path in enumerate(paths):

        # Remove the path
        if os.path.exists(path):
            if os.path.isdir(path):
                logger.out('Removing directory: {}'.format(path))
                if not dry_run:
                    shutil.rmtree(path)
            else:
                logger.out('     Removing file: {}'.format(path))
                if not dry_run:
                    os.remove(path)
        else:
            logger.err('Warning: File {} could not be deleted because it does not exist.'.format(path))

        # Report progress
        event_queue.put(ExecutionEvent(
            'STAGE_PROGRESS',
            job_id=logger.job_id,
            stage_label=logger.stage_label,
            percent=str(index / float(len(paths)))))

    return 0



# Copyright 2014 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import argparse
import logging
import os
import sys

try:
    from catkin_pkg.packages import find_packages
    from catkin_pkg.topological_order import topological_order_packages
except ImportError as e:
    sys.exit(
        'ImportError: "from catkin_pkg.topological_order import '
        'topological_order" failed: %s\nMake sure that you have installed '
        '"catkin_pkg", and that it is up to date and on the PYTHONPATH.' % e
    )

from catkin_pkg.package import InvalidPackage
from catkin_pkg.tool_detection import get_previous_tool_used_on_the_space
from catkin_pkg.tool_detection import mark_space_as_built_by

from ckx_tools.argument_parsing import add_context_args
from ckx_tools.argument_parsing import add_cmake_and_make_and_catkin_make_args
from ckx_tools.argument_parsing import configure_make_args

from ckx_tools.common import get_build_type
from ckx_tools.common import getcwd
from ckx_tools.common import is_tty
from ckx_tools.common import log
from ckx_tools.common import find_enclosing_package
from ckx_tools.common import format_env_dict

from ckx_tools.context import Context

import ckx_tools.metadata as metadata
import ckx_tools.execution.job_server as job_server

from ckx_tools.jobs.utils import get_env_loader

from ckx_tools.metadata import find_enclosing_workspace
from ckx_tools.metadata import get_metadata
from ckx_tools.metadata import update_metadata

from ckx_tools.resultspace import load_resultspace_environment

from ckx_tools.terminal_color import set_color

from .color import clr

from .build import build_isolated_workspace
from .build import determine_packages_to_be_built
from .build import verify_start_with_option

#
# Begin Hack
#

# TODO(wjwwood): remove this, once it is no longer needed.
# argparse may not support mutually exclusive groups inside other groups, see:
#   http://bugs.python.org/issue10680

# Backup the original constructor
backup__ArgumentGroup___init__ = argparse._ArgumentGroup.__init__


# Make a new constructor with the fix
def fixed__ArgumentGroup___init__(self, container, title=None, description=None, **kwargs):
    backup__ArgumentGroup___init__(self, container, title, description, **kwargs)
    # Make sure this line is run, maybe redundant on versions which already have it
    self._mutually_exclusive_groups = container._mutually_exclusive_groups


# Monkey patch in the fixed constructor
argparse._ArgumentGroup.__init__ = fixed__ArgumentGroup___init__

#
# End Hack
#

##############################################################################
# Entry Point API
##############################################################################


def prepare_arguments(parser):
    parser.description = """\
Build one or more packages in a catkin workspace.
This invokes `CMake`, `make`, and optionally `make install` for either all
or the specified packages in a catkin workspace.

Arguments passed to this verb can temporarily override persistent options
stored in the catkin profile config. If you want to save these options, use
the --save-config argument. To see the current config, use the
`catkin config` command.\
"""

    # Workspace / profile args
    add_context_args(parser)

    common_group = parser.add_argument_group('Common Options', 'Most frequently used options.')
    add = common_group.add_argument
    add('--verbose', '-v', action='store_true', default=False,
        help='Print output from commands in ordered blocks once the command finishes.')
    add('--force-cmake', action='store_true', default=None,
        help='Runs cmake explicitly for each catkin package.')
    add('--pre-clean', action='store_true', default=None,
        help='Runs `make clean` before building each package.')
    add('--install', action='store_true', dest='install_only_this_time', default=None,
        help='Enable installation after the build has finished (overrides config settings)')
    add('packages', metavar='PKGNAME', nargs='*',
        help='Workspace packages to build, package dependencies are built as well unless --no-deps is used. '
             'If no packages are given, then all the packages are built.')
    # test_group = parser.add_mutually_exclusive_group()
    # add = test_group.add_argument
    # add('-t', '--tests', action='store_true', help='Make tests [false]')
    # add('-r', '--run-tests', action='store_true', help='Run tests (does not build them) [false]')

    # What packages to build
    pkg_group = parser.add_argument_group('Advanced Package Control', 'Control which packages get built.')
    add = pkg_group.add_argument
    add('--this', dest='build_this', action='store_true', default=False,
        help='Build the package containing the current working directory.')
    add('--no-deps', action='store_true', default=False,
        help='Only build specified packages, not their dependencies.')
    add('--unbuilt', action='store_true', default=False,
        help='Build packages which have yet to be built.')

    start_with_group = pkg_group.add_mutually_exclusive_group()
    add = start_with_group.add_argument
    add('--start-with', metavar='PKGNAME', type=str,
        help='Build a given package and those which depend on it, skipping any before it.')
    add('--start-with-this', action='store_true', default=False,
        help='Similar to --start-with, starting with the package containing the current directory.')
    add = pkg_group.add_argument
    add('--continue-on-failure', '-c', action='store_true', default=False,
        help='Try to continue building packages whose dependencies built successfully even if some other requested '
             'packages fail to build.')

    # Build options
    build_group = parser.add_argument_group('Advanced Build Options', 'Control the build behavior.')
    add = build_group.add_argument
    add('--strip', action='store_true', help='Strips binaries, only valid with --install')
    add('--no-install-lock', action='store_true', default=None,
        help='Prevents serialization of the install steps, which is on by default to prevent file install collisions')

    config_group = parser.add_argument_group('Advanced Configuration', 'Parameters for the underlying build system.')
    add = config_group.add_argument
    add('--save-config', action='store_true', default=False,
        help='Save any configuration options in this section for the next build invocation.')
    add_cmake_and_make_and_catkin_make_args(config_group)

    # Behavior
    behavior_group = parser.add_argument_group('Advanced Interface Options', 'The behavior of the command-line interface.')
    add = behavior_group.add_argument
    add('--interleave-output', '-i', action='store_true', default=False,
        help='Prevents ordering of command output when multiple commands are running at the same time.')
    add('--no-status', action='store_true', default=False,
        help='Suppresses status line, useful in situations where carriage return is not properly supported.')
    add('--summarize', '--summary', '-s', action='store_true', default=None,
        help='Adds a build summary to the end of a build; defaults to on with --continue-on-failure, off otherwise')
    add('--no-summarize', '--no-summary', action='store_false', dest='summarize',
        help='Explicitly disable the end of build summary')
    add('--override-build-tool-check', action='store_true', default=False,
        help='use to override failure due to using differnt build tools on the same workspace.')

    # Introspection
    introspection_group = parser.add_argument_group('Introspection', 'Build environment analysis')
    add = introspection_group.add_argument
    add('--dry-run', '-n', action='store_true', default=False,
        help='List the packages which will be built with the given arguments without building them.')
    add('--get-env', dest='get_env', metavar='PKGNAME', nargs=1,
        help='Print the environment in which PKGNAME is built to stdout.')

    # Experimental args
    add('--mem-limit', default=None, help=argparse.SUPPRESS)
    # Advanced args
    add('--develdebug', metavar='LEVEL', default=None, help=argparse.SUPPRESS)

    def status_rate_type(rate):
        rate = float(rate)
        if rate < 0:
            raise argparse.ArgumentTypeError("must be greater than or equal to zero.")
        return rate

    add('--limit-status-rate', '--status-rate', type=status_rate_type, default=10.0,
        help='Limit the update rate of the status bar to this frequency. Zero means unlimited. '
             'Must be positive, default is 10 Hz.')
    add('--no-notify', action='store_true', default=False,
        help='Suppresses system pop-up notification.')

    return parser


def main(opts):

    # Check for develdebug mode
    if opts.develdebug is not None:
        os.environ['TROLLIUSDEBUG'] = opts.develdebug.lower()
        logging.basicConfig(level=opts.develdebug.upper())

    # Set color options
    opts.force_color = os.environ.get('CATKIN_TOOLS_FORCE_COLOR', opts.force_color)
    if (opts.force_color or is_tty(sys.stdout)) and not opts.no_color:
        set_color(True)
    else:
        set_color(False)

    # Context-aware args
    if opts.build_this or opts.start_with_this:
        # Determine the enclosing package
        try:
            ws_path = find_enclosing_workspace(getcwd())
            # Suppress warnings since this won't necessaraly find all packages
            # in the workspace (it stops when it finds one package), and
            # relying on it for warnings could mislead people.
            this_package = find_enclosing_package(
                search_start_path=getcwd(),
                ws_path=ws_path,
                warnings=[])
        except (InvalidPackage, RuntimeError):
            this_package = None

        # Handle context-based package building
        if opts.build_this:
            if this_package:
                opts.packages += [this_package]
            else:
                sys.exit(
                    "[build] Error: In order to use --this, the current directory must be part of a catkin package.")

        # If --start--with was used without any packages and --this was specified, start with this package
        if opts.start_with_this:
            if this_package:
                opts.start_with = this_package
            else:
                sys.exit(
                    "[build] Error: In order to use --this, the current directory must be part of a catkin package.")

    if opts.no_deps and not opts.packages and not opts.unbuilt:
        sys.exit(clr("[build] @!@{rf}Error:@| With --no-deps, you must specify packages to build."))

    # Are we in a parallel build profile? If so, prefer that over the active one
    if not opts.profile:
        enclosing_profile = metadata.find_enclosing_profile(os.getcwdu(), opts.workspace)
        if enclosing_profile:
            log(clr("@!\nInfo: in a parallel build folder, prefer this profile [%s]\n@|" % enclosing_profile))
            opts.profile = enclosing_profile

    # Load the context
    ctx = Context.load(opts.workspace, opts.profile, opts, append=True)

    # Initialize the build configuration
    make_args, makeflags, cli_flags, jobserver = configure_make_args(
        ctx.make_args, ctx.jobs_args, ctx.use_internal_make_jobserver)

    # Set the jobserver memory limit
    if jobserver and opts.mem_limit:
        log(clr("@!@{pf}EXPERIMENTAL: limit memory to '%s'@|" % str(opts.mem_limit)))
        # At this point psuitl will be required, check for it and bail out if not set
        try:
            import psutil  # noqa
        except ImportError as exc:
            log("Could not import psutil, but psutil is required when using --mem-limit.")
            log("Please either install psutil or avoid using --mem-limit.")
            sys.exit("Exception: {0}".format(exc))
        job_server.set_max_mem(opts.mem_limit)

    ctx.make_args = make_args

    # Load the environment of the workspace to extend
    if ctx.underlays is not None:
        load_resultspace_environment(ctx.underlays)

    # Check if the context is valid before writing any metadata
    if not ctx.source_space_exists():
        sys.exit(clr("[build] @!@{rf}Error:@| Unable to find source space `%s`") % ctx.source_space_abs)

    # ensure the build space was previously built by ckx_tools
    previous_tool = get_previous_tool_used_on_the_space(ctx.build_space_abs)
    if previous_tool is not None and previous_tool != 'catkin build':
        if opts.override_build_tool_check:
            log(clr(
                "@{yf}Warning: build space at '%s' was previously built by '%s', "
                "but --override-build-tool-check was passed so continuing anyways."
                % (ctx.build_space_abs, previous_tool)))
        else:
            sys.exit(clr(
                "@{rf}The build space at '%s' was previously built by '%s'. "
                "Please remove the build space or pick a different build space."
                % (ctx.build_space_abs, previous_tool)))
    # the build space will be marked as catkin build's if dry run doesn't return

    # ensure the devel space was previously built by ckx_tools
    previous_tool = get_previous_tool_used_on_the_space(ctx.devel_space_abs)
    if previous_tool is not None and previous_tool != 'catkin build':
        if opts.override_build_tool_check:
            log(clr(
                "@{yf}Warning: devel space at '%s' was previously built by '%s', "
                "but --override-build-tool-check was passed so continuing anyways."
                % (ctx.devel_space_abs, previous_tool)))
        else:
            sys.exit(clr(
                "@{rf}The devel space at '%s' was previously built by '%s'. "
                "Please remove the devel space or pick a different devel space."
                % (ctx.devel_space_abs, previous_tool)))
    # the devel space will be marked as catkin build's if dry run doesn't return

    # Display list and leave the file system untouched
    if opts.dry_run:
        # TODO: Add unbuilt
        dry_run(ctx, opts.packages, opts.no_deps, opts.start_with)
        return 0

    # Print the build environment for a given package and leave the filesystem untouched
    if opts.get_env:
        return print_build_env(ctx, opts.get_env[0])

    # Now mark the build and devel spaces as catkin build's since dry run didn't return.
    mark_space_as_built_by(ctx.build_space_abs, 'catkin build')
    mark_space_as_built_by(ctx.devel_space_abs, 'catkin build')

    # Get the last build context
    build_metadata = get_metadata(ctx.workspace, ctx.profile, 'build')

    # Force cmake if the CMake arguments have changed
    if build_metadata.get('cmake_args') != ctx.cmake_args:
        opts.force_cmake = True

    # Check the devel layout compatibility
    last_devel_layout = build_metadata.get('devel_layout', ctx.devel_layout)
    if last_devel_layout != ctx.devel_layout:
        sys.exit(clr(
            "@{rf}@!Error:@|@{rf} The current devel space layout, `{}`,"
            "is incompatible with the configured layout, `{}`.@|").format(
            last_devel_layout, ctx.devel_layout))

    # Check if some other verb has changed the workspace in such a way that it needs to be forced
    if build_metadata.get('needs_force', False):
        opts.force_cmake = True
        update_metadata(ctx.workspace, ctx.profile, 'build', {'needs_force': False})

    # Always save the last context under the build verb
    update_metadata(ctx.workspace, ctx.profile, 'build', ctx.get_stored_dict())

    # Save the context as the configuration
    if opts.save_config:
        Context.save(ctx)

    # Get parallel toplevel jobs
    try:
        parallel_jobs = int(opts.parallel_jobs)
    except TypeError:
        parallel_jobs = None

    # Set VERBOSE environment variable
    if opts.verbose:
        os.environ['VERBOSE'] = '1'

    if opts.install_only_this_time:
        ctx.install = opts.install_only_this_time  # override

    return build_isolated_workspace(
        ctx,
        packages=opts.packages,
        start_with=opts.start_with,
        no_deps=opts.no_deps,
        unbuilt=opts.unbuilt,
        n_jobs=parallel_jobs,
        force_cmake=opts.force_cmake,
        pre_clean=opts.pre_clean,
        force_color=opts.force_color,
        quiet=not opts.verbose,
        interleave_output=opts.interleave_output,
        no_status=opts.no_status,
        limit_status_rate=opts.limit_status_rate,
        lock_install=not opts.no_install_lock,
        no_notify=opts.no_notify,
        continue_on_failure=opts.continue_on_failure,
        summarize_build=opts.summarize  # Can be True, False, or None
    )

##############################################################################
# Helpers
##############################################################################


def dry_run(context, packages, no_deps, start_with):
    # Print Summary
    log(context.summary())
    # Get all the packages in the context source space
    # Suppress warnings since this is a utility function
    workspace_packages = find_packages(context.source_space_abs, exclude_subspaces=True, warnings=[])
    # Find list of packages in the workspace
    packages_to_be_built, packages_to_be_built_deps, all_packages = determine_packages_to_be_built(
        packages, context, workspace_packages)
    # Assert start_with package is in the workspace
    verify_start_with_option(start_with, packages, all_packages, packages_to_be_built + packages_to_be_built_deps)
    if not no_deps:
        # Extend packages to be built to include their deps
        packages_to_be_built.extend(packages_to_be_built_deps)
        # Also resort
        packages_to_be_built = topological_order_packages(dict(packages_to_be_built))
    # Print packages
    log("Packages to be built:")
    max_name_len = str(max([len(pkg.name) for pth, pkg in packages_to_be_built]))
    prefix = clr('@{pf}' + ('------ ' if start_with else '- ') + '@|')
    for pkg_path, pkg in packages_to_be_built:
        build_type = get_build_type(pkg)
        if build_type == 'catkin' and 'metapackage' in [e.tagname for e in pkg.exports]:
            build_type = 'metapackage'
        if start_with and pkg.name == start_with:
            start_with = None
        log(clr("{prefix}@{cf}{name:<" + max_name_len + "}@| (@{yf}{build_type}@|)")
            .format(prefix=clr('@!@{kf}(skip)@| ') if start_with else prefix, name=pkg.name, build_type=build_type))
    log("Total packages: " + str(len(packages_to_be_built)))


def print_build_env(context, package_name):
    workspace_packages = find_packages(context.source_space_abs, exclude_subspaces=True, warnings=[])
    # Load the environment used by this package for building
    for pth, pkg in workspace_packages.items():
        if pkg.name == package_name:
            environ = get_env_loader(pkg, context)(os.environ)
            print(format_env_dict(environ))
            return 0
    print('[build] Error: Package `{}` not in workspace.'.format(package_name),
          file=sys.stderr)
    return 1




#!/usr/bin/env python
#
# License: BSD
#   https://raw.github.com/stonier/ckx_tools/devel/LICENSE
#
##############################################################################
# Documentation
##############################################################################

"""
Implementation of the 'ckx rosdoc' verb.
"""

##############################################################################
# Imports
##############################################################################

import os
import multiprocessing

import subprocess
import traceback

try:
    from queue import Queue # Python3
except ImportError:
    from Queue import Queue # Python2

import catkin_pkg.packages as catkin_packages
import ckx_tools.argument_parsing as argument_parsing
import ckx_tools.common as common
import ckx_tools.execution.job_server as job_server
import ckx_tools.metadata as metadata

from ckx_tools.context import Context
from ckx_tools.execution.executor import execute_jobs
from ckx_tools.execution.executor import run_until_complete
from ckx_tools.execution.jobs import Job
from ckx_tools.execution.stages import CommandStage
from ckx_tools.execution.controllers import ConsoleStatusController
from ckx_tools.terminal_color import ColorMapper
from ckx_tools.utils import which

from . import templates

color_mapper = ColorMapper()
clr = color_mapper.clr

##############################################################################
# Entry Point API
##############################################################################

def help_string():
    instructions = clr("@!Examples@|\n\n  \
  @{cf}ckx rosdoc@| : @{yf}doc all packages in the source workspace@|\n  \
  @{cf}ckx rosdoc ecl_build ecl_time_lite@| : @{yf}doc only the listed packages@|\n\n  \
")
    return instructions

def prepare_arguments(parser):
    parser.epilog = help_string()
    argument_parsing.add_context_args(parser) # workspace / profile args
    add = parser.add_argument
    add('packages', nargs='*', help="workspace packages to doc, if none are given, then all are generated")
    add('--verbose', '-v', action='store_true', default=False, help='Print output from doc generators in their full glory.')
    return parser


def main(opts):
    # Do we have rosdoc_lite?
    if not which(DOC_PROGRAM):
        print(clr("[rosdoc] @!@{rf}Error: {0}' is not available.@|").format(DOC_PROGRAM))
        print(clr("[rosdoc] @!@{rf}Error: please make sure '{0}' is installed and in your path@|").format(DOC_PROGRAM))
        return 1

    # Are we in a parallel build profile? If so, prefer that over the active one
    if not opts.profile:
        enclosing_profile = metadata.find_enclosing_profile(os.getcwdu(), opts.workspace)
        if enclosing_profile:
            print(clr("@!\nInfo: in a parallel build folder, prefer this profile [%s]\n@|" % enclosing_profile))
            opts.profile = enclosing_profile

    # Load the context
    context = Context.load(opts.workspace, opts.profile, opts)

    if not os.path.exists(context.doc_space_abs):
        os.mkdir(context.doc_space_abs)

    # Initialize job server
    job_server.initialize(
        max_jobs=None,  # let the job server decide
        max_load=None,
        gnu_make_enabled=False)

    # List of packages, curtail if restricted list is specified
    packages = catkin_packages.find_packages(context.source_space_abs, exclude_subspaces=True)
    if opts.packages:
        packages = {k: v for (k, v) in packages.iteritems() if v.name in opts.packages}

    # List up packages with its absolute path
    packages_by_name = {p.name: os.path.join(context.source_space_abs, path) for path, p in packages.iteritems()}

    # doc_output = {}
    stages = []
    print('\nGenerating documents in {0}\n'.format(context.doc_space_abs))
    for name, path in packages_by_name.items():
        print(' - {0}'.format(name))
        stages.append(generate_doc_stage(name, path, context.doc_space_abs, context.workspace))
        # output = generate_doc(name, path, context.doc_space_abs)
        # doc_output[name] = output
    job = Job(
        jid="rosdoc",  # unique job identifier
        deps=[],
        env_loader=load_env,  # get_env_loader(package, context),
        stages=stages)
    event_queue = Queue()
    status_thread = ConsoleStatusController(
        'rosdoc',
        job_labels=['rosdoc', 'lite'],
        jobs=[job],
        max_toplevel_jobs=multiprocessing.cpu_count(),
        available_jobs=[],  # [pkg.name for _, pkg in context.packages],
        whitelisted_jobs= [],
        blacklisted_jobs=[],
        event_queue=event_queue,
        show_buffered_stdout=opts.verbose,
    )
    status_thread.start()
    # Block while running N jobs asynchronously
    try:
        unused_all_succeeded = run_until_complete(execute_jobs(
            'rosdoc',
            jobs=[job],
            locks={},
            event_queue=event_queue,
            log_path=context.log_space_abs,
            max_toplevel_jobs=multiprocessing.cpu_count()
            ))
    except Exception:
        status_thread.keep_running = False
        status_thread.join(1.0)
        common.wide_log(str(traceback.format_exc()))
    status_thread.join(1.0)

    generates_index_page(context.doc_space_abs, packages_by_name.keys())

#     print('\nDocument generation result. 0 may mean error. But it is fine most of time\n')
#     for name, err in doc_output.items():
#         print(" - {0} : {1}".format(name, str(err)))
#     print("")
    return 0

##############################################################################
# Helpers
##############################################################################

DOC_PROGRAM = 'rosdoc_lite'


def generate_doc(name, pkg_path, doc_path):
    document_path = doc_path + '/' + name
    args = [DOC_PROGRAM, '-o', document_path, pkg_path]
    output = subprocess.call(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return output

def generate_doc_stage(name, pkg_path, doc_path, workspace_path):
    document_path = doc_path + '/' + name
    cmd = [DOC_PROGRAM, '-o', document_path, pkg_path]
    return CommandStage(
        label=name,
        cmd=cmd,
        cwd=workspace_path,
        occupy_job=True
        )


def output(fd, html):
    for h in html:
        os.write(fd, h)
        os.write(fd, '\n')


def generates_index_page(doc_path, pkg_names):

    index_page = doc_path + '/index.html'
    fd = os.open(index_page, os.O_RDWR | os.O_CREAT)

    output(fd, templates.html_header)

    for pkg in pkg_names:
        link_html = '  <a href="' + pkg + '/html/index.html">' + pkg + '</a><br/>'
        output(fd, [link_html])

    output(fd, templates.html_footer)
    os.close(fd)

def load_env(base_env):
    """
    @todo need a better environment than this so it can actually find
    rosdoc_lite without having to pre-load the environment from outside.

    :param base_env: usually this is just os.environ as is
    """
    return base_env





# Copyright 2017 Daniel Stonier
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

##############################################################################
# Documentation
##############################################################################

"""
Supporting utilities to instantation cmake cache files, toolchains and
script launchers.
"""

##############################################################################
# Imports
##############################################################################

import os
import shutil
import stat  # file permissions

import ckx_tools.config

from . import color_console as console

##############################################################################
# Methods
##############################################################################

config_cache_filename = os.path.join(os.path.dirname(__file__), 'resources', 'cmake', 'config.cmake', )
overrides_filename = os.path.join(os.path.dirname(__file__), 'resources', 'cmake', 'overrides.cmake')

templates_home = os.path.join(os.path.dirname(__file__), 'resources', 'templates')
toolchains_home = os.path.join(os.path.dirname(__file__), 'resources', 'toolchains')
custom_toolchains_home = os.path.join(ckx_tools.config.home(), 'toolchains')
platforms_home = os.path.join(os.path.dirname(__file__), 'resources', 'platforms')
custom_platforms_home = os.path.join(ckx_tools.config.home(), 'platforms')


def file_to_string(filename):
    f = open(filename, 'r')
    try:
        t = f.read()
    finally:
        f.close()
    return t

def fill_in_session_scripts(template, name, cwd):
    return template % locals()

def instantiate_or_update_session_scripts(name, build_root_dir, workspace_dir):
    (_, _, filenames) = os.walk(templates_home).next()
    for filename in filenames:
        if not os.path.isfile(os.path.join(build_root_dir, filename)):
            raw_contents = file_to_string(os.path.join(templates_home, filename))
            contents = fill_in_session_scripts(raw_contents, name, workspace_dir)
            try:
                f = open(os.path.join(build_root_dir, filename), 'w')
                f.write(contents.encode('utf-8'))
            finally:
                os.fchmod(f.fileno(), stat.S_IRWXU | stat.S_IRWXG)
                f.close()

def fill_in_config_cmake(template, config_workspace, config_build_type, config_doc_prefix, config_underlays, config_override_file):
    return template % locals()

def instantiate_or_update_config_environment(
        name,
        workspace_dir,
        build_root_dir,
        platform,
        toolchain,
        doc_prefix,
        underlays
    ):
    """
    Instatiate if not already instatiated, or validate/update configuration
    files - scripts, cmake cache and if used, the toolchain module.

    :param str name: the build profile name
    :param str workspace_dir:
    :param str build_root_dir:
    :param str platform: unique identifier used to lookup the platform library
    :param str toolchain: unique identifier used to lookup the toolchain library
    :param str doc_prefix:
    :param list underlays: semi-colon separated list of underlay roots

    :raises RuntimeError: if it fails to setup anything
    """

    if not os.path.exists(build_root_dir):
        os.makedirs(build_root_dir)
    instantiate_or_update_session_scripts(
        name=os.path.basename(workspace_dir) + "_" + name,
        build_root_dir=build_root_dir,
        workspace_dir=workspace_dir
    )
    # this can throw a RuntimeError if the platform configuration is not found
    instantiate_or_update_config_cmake(
        platform,
        workspace_dir,
        build_root_dir,
        doc_prefix,
        underlays
    )
    # this can throw a RuntimeError if the toolchain is not found
    instantiate_or_udpate_toolchain_module(
        toolchain,
        build_root_dir
    )


def instantiate_or_update_config_cmake(platform_name, workspace_dir, build_root_dir, config_doc_prefix, config_underlays):
    '''
    Copy the cache configuration template to the build path.

    :param platform_name: name of the platform configuration to bring to the cache
    :param build_root_dir: location of the build directory to save the config.cmake file.

    :raises RuntimeError: if it fails to setup anything
    '''
    # this one can throw a RuntimeError
    platform_content = get_platform_content(platform_name)
    config_build_type = "RelWithDebInfo"
    if os.path.isfile(os.path.join(build_root_dir, 'config.cmake')):
        # TODO : validate an existing cache file to make sure it hasn't been unintentially put in a conflicting state by the user
        return
    raw_content = file_to_string(config_cache_filename)
    contents = fill_in_config_cmake(raw_content, workspace_dir, config_build_type, config_doc_prefix, config_underlays, overrides_filename)
    config_cmake_file = os.path.join(build_root_dir, "config.cmake")
    try:
        f = open(config_cmake_file, 'w')
        f.write(platform_content.encode('utf-8'))
        f.write(contents.encode('utf-8'))
    finally:
        f.close()

def instantiate_or_udpate_toolchain_module(toolchain, build_root_dir):
    """
    Instantate with a toolchain module, or remove/replace it if the configuration
    is different to the file that is present.

    :param str toolchain: string identifier representing a toolchain in the library
    :param str build_root_dir:

    :raises RuntimeError: if it can't find the toolchain module in the library
    """
    if toolchain:
        # TODO assert the file is there and the right toolchain and if not, proceed
        tmp_list = toolchain.split('/')
        if len(tmp_list) != 2:
            raise RuntimeError("Toolchain specification invalid, must be <family>/<tuple> [%s]" % toolchain)
        family = tmp_list[0]
        toolchain_tuple = tmp_list[1]
        toolchains = get_toolchains_or_platforms(toolchains_home)
        custom_toolchains = get_toolchains_or_platforms(custom_toolchains_home)
        if family not in toolchains and family not in custom_toolchains:
            raise RuntimeError("No toolchains available for family %s" % family)
        if family in custom_toolchains and toolchain_tuple in custom_toolchains[family]:
            toolchain_file = os.path.join(custom_toolchains_home, family, toolchain_tuple + ".cmake")
        elif family in toolchains and toolchain_tuple in toolchains[family]:
            toolchain_file = os.path.join(toolchains_home, family, toolchain_tuple + ".cmake")
        else:
            raise RuntimeError("Platform %s for family %s not available." % (family, toolchain_tuple))
        if os.path.isfile(toolchain_file):
            shutil.copy(toolchain_file, os.path.join(build_root_dir, "toolchain.cmake"))
        else:
            raise RuntimeError("Toolchain module not available [%s]" % toolchain_file)
    else:
        toolchain_filename = os.path.join(build_root_dir, "toolchain.cmake")
        if os.path.isfile(toolchain_filename):
            os.remove(toolchain_filename)


def get_toolchains_or_platforms(base_path):
    '''
      Does a look up in the path for either toolchain or platform files.
      @param base_path : start of the search (ends in either toolchains or platforms).
    '''
    d = {}
    for (dirpath, unused_dirname, filenames) in os.walk(base_path):
        if dirpath != base_path:  # skip the root, only find files in subfolders
            family = os.path.basename(dirpath)
            d[family] = []
            for filename in filenames:
                d[family].append(os.path.splitext(filename)[0])  # leave off the .cmake extension
    return d

def get_platform_content(platform_name):
    """
    Retrieve cmake platform configuration from the library.

    :raises RuntimeError: if it can't find the platform configuration file.
    """
    platform_content = ""
    if platform_name:
        tmp_list = platform_name.split('/')
        if len(tmp_list) != 2:
            raise RuntimeError("Platform specification invalid, must be <family>/<platform type> [%s]" % platform_name)
        family = tmp_list[0]
        platform = tmp_list[1]
        platforms = get_toolchains_or_platforms(platforms_home)
        custom_platforms = get_toolchains_or_platforms(custom_platforms_home)
        if family not in platforms and family not in custom_platforms:
            raise RuntimeError("No platforms available for family %s" % family)
        if family in custom_platforms and platform in custom_platforms[family]:
            platform_file = os.path.join(custom_platforms_home, family, platform + ".cmake")
        elif family in platforms and platform in platforms[family]:
            platform_file = os.path.join(platforms_home, family, platform + ".cmake")
        else:
            raise RuntimeError("platform %s for family %s not available." % (family, platform))
    else:
        platform_file = os.path.join(platforms_home, 'default.cmake')

    if os.path.isfile(platform_file):
        f = open(platform_file, 'r')
        try:
            platform_content = f.read()
        finally:
            f.close()
    else:
        raise RuntimeError("platform configuration not available [%s]" % platform_file)
    return platform_content

def list_toolchains():
    console.pretty_println("\n******************************** Toolchain List **********************************", console.bold)
    toolchains = get_toolchains_or_platforms(toolchains_home)
    print(console.bold + "Official " + console.reset + "({0})".format(toolchains_home))
    for family in toolchains:
        for platform in toolchains[family]:
            console.pretty_print(" -- %s/" % family, console.cyan)
            console.pretty_println("%s" % platform, console.yellow)
    print(console.bold + "Custom " + console.reset + "({0})".format(custom_toolchains_home))
    toolchains = get_toolchains_or_platforms(custom_toolchains_home)
    if toolchains:
        for family in toolchains:
            for platform in toolchains[family]:
                console.pretty_print(" -- %s/" % family, console.cyan)
                console.pretty_println("%s" % platform, console.yellow)
    else:
        console.pretty_println(" -- ", console.cyan)
    console.pretty_println("**********************************************************************************\n", console.bold)


def list_platforms():
    console.pretty_println("\n********************************* Platform List **********************************", console.bold)
    platforms = get_toolchains_or_platforms(platforms_home)
    print(console.bold + "Official " + console.reset + "({0})".format(platforms_home))
    for family in platforms:
        for platform in platforms[family]:
            console.pretty_print(" -- %s/" % family, console.cyan)
            console.pretty_println("%s" % platform, console.yellow)
    print(console.bold + "Custom " + console.reset + "({0})".format(custom_platforms_home))
    platforms = get_toolchains_or_platforms(custom_platforms_home)
    if platforms:
        for family in platforms:
            for platform in platforms[family]:
                console.pretty_print(" -- %s/" % family, console.cyan)
                console.pretty_println("%s" % platform, console.yellow)
    else:
        console.pretty_println(" -- ", console.cyan)
    console.pretty_println("**********************************************************************************\n", console.bold)




import argparse
from distutils import log
import os
import site
from stat import ST_MODE
import sys

from setuptools import find_packages
from setuptools import setup
from setuptools.command.install import install

# Setup installation dependencies, removing some so they
# can build on the ppa
install_requires = [
    'catkin-pkg > 0.2.9',
    'setuptools',
    'PyYAML',
    'osrf-pycommon > 0.1.1',
    'rospkg',
    'vci',
    'vcstool'
]
if sys.version_info[0] == 2 and sys.version_info[1] <= 6:
    install_requires.append('argparse')

# Figure out the resources that need to be installed
this_dir = os.path.abspath(os.path.dirname(__file__))
osx_resources_path = os.path.join(
    this_dir,
    'ckx_tools',
    'notifications',
    'resources',
    'osx',
    'catkin build.app')
osx_notification_resources = [os.path.join(dp, f)
                              for dp, dn, fn in os.walk(osx_resources_path)
                              for f in fn]
src_path = os.path.join(this_dir, 'ckx_tools')
osx_notification_resources = [os.path.relpath(x, src_path)
                              for x in osx_notification_resources]


def _resolve_prefix(prefix, type):
    osx_system_prefix = '/System/Library/Frameworks/Python.framework/Versions'
    if type == 'man':
        if prefix == '/usr':
            return '/usr/share'
        if sys.prefix.startswith(osx_system_prefix):
            return '/usr/share'
    elif type == 'bash_comp':
        if prefix == '/usr':
            return '/'
        if sys.prefix.startswith(osx_system_prefix):
            return '/'
    elif type == 'zsh_comp':
        if sys.prefix.startswith(osx_system_prefix):
            return '/usr/local'
    else:
        raise ValueError('not supported type')
    return prefix


def get_data_files(prefix):
    data_files = []

    # Bash completion
    bash_comp_dest = os.path.join(_resolve_prefix(prefix, 'bash_comp'),
                                  'etc/bash_completion.d')
    data_files.append((bash_comp_dest,
                       ['completion/ckx_tools-completion.bash']))

    # Zsh completion
    zsh_comp_dest = os.path.join(_resolve_prefix(prefix, 'zsh_comp'),
                                 'share/zsh/site-functions')
    data_files.append((zsh_comp_dest, ['completion/_ckx']))
    return data_files


class PermissiveInstall(install):

    def run(self):
        install.run(self)
        if os.name == 'posix':
            for file in self.get_outputs():
                # all installed files should be readable for anybody
                mode = ((os.stat(file)[ST_MODE]) | 0o444) & 0o7777
                log.info("changing permissions of %s to %o" % (file, mode))
                os.chmod(file, mode)

        # Provide information about bash completion after default install.
        if (sys.platform.startswith("linux") and
                self.install_data == "/usr/local"):
            log.info("""
----------------------------------------------------------------
To enable tab completion, add the following to your '~/.bashrc':

  source {0}

----------------------------------------------------------------
""".format(os.path.join(self.install_data,
                        'etc/bash_completion.d',
                        'ckx_tools-completion.bash')))

parser = argparse.ArgumentParser(add_help=False)
prefix_group = parser.add_mutually_exclusive_group()
prefix_group.add_argument('--user', '--home', action='store_true')
prefix_group.add_argument('--prefix', default=None)

opts, _ = parser.parse_known_args(sys.argv)
userbase = site.getuserbase() if opts.user else None
prefix = userbase or opts.prefix or sys.prefix

setup(
    name='ckx_tools',
    version='0.5.15',
    packages=find_packages(exclude=['tests*', 'docs*']),  # ['ckx_tools'],  # find_packages(exclude=['tests', 'docs']), <-- broken, it picks up sub-packages, e.g. tests.unit
    package_data={
        'ckx_tools': [
            'notifications/resources/linux/*',
            'verbs/ckx_config/resources/cmake/*',
            'verbs/ckx_config/resources/templates/*',
            'verbs/ckx_config/resources/toolchains/buildroot/*',
            'verbs/ckx_config/resources/toolchains/ubuntu/*',
            'verbs/ckx_config/resources/toolchains/code_sourcery/*',
            'verbs/ckx_config/resources/toolchains/nexell/*',
            'verbs/ckx_config/resources/platforms/default.cmake',
            'verbs/ckx_config/resources/platforms/generic/*',
            'verbs/ckx_config/resources/platforms/arm/*',
            'verbs/ckx_config/resources/platforms/intel/*',
            'verbs/ckx_shell_verbs.bash',
            'docs/examples',
        ] + osx_notification_resources
    },
    data_files=get_data_files(prefix),
    install_requires=install_requires,
    author='Daniel Stonier',
    author_email='d.stonier@gmail.com',
    maintainer='Daniel Stonier',
    maintainer_email='d.stonier@gmail.com',
    url='http://ckx-tools.readthedocs.org/',
    keywords=['catkin'],
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
    ],
    description="'X' tools for working with catkin.",
    long_description="Provides command line tools for working with catkin.",
    license='Apache 2.0',
    # test_suite='tests',
    entry_points={
        'console_scripts': [
            'ckx = ckx_tools.commands.catkin:main',
        ],
        'ckx_tools.commands.catkin.verbs': [
            'build = ckx_tools.verbs.ckx_build:description',
            'clean = ckx_tools.verbs.ckx_clean:description',
            'config = ckx_tools.verbs.ckx_config:description',
            'create = ckx_tools.verbs.ckx_create:description',
            'env = ckx_tools.verbs.ckx_env:description',
            'list = ckx_tools.verbs.ckx_list:description',
            'locate = ckx_tools.verbs.ckx_locate:description',
            'profile = ckx_tools.verbs.ckx_profile:description',
            'rosdep = ckx_tools.verbs.ckx_rosdep:description',
            'rosdoc = ckx_tools.verbs.ckx_rosdoc:description',
            'ws = ckx_tools.verbs.ckx_ws:description',
        ],
        'ckx_tools.jobs': [
            'catkin = ckx_tools.jobs.catkin:description',
            'cmake = ckx_tools.jobs.cmake:description',
        ],
    },
    cmdclass={'install': PermissiveInstall},
)



# Copyright 2014 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import os
import pkg_resources
import shutil
import yaml

from . import common

METADATA_DIR_NAME = '.ckx_tools'

METADATA_README_TEXT = """\
# Catkin Tools Metadata

This directory was generated by ckx_tools and it contains persistent
configuration information used by the `catkin` command and its sub-commands.

Each subdirectory of the `profiles` directory contains a set of persistent
configuration options for separate profiles. The default profile is called
`default`. If another profile is desired, it can be described in the
`profiles.yaml` file in this directory.

Please see the ckx_tools documentation before editing any files in this
directory. Most actions can be performed with the `catkin` command-line
program.
"""

PROFILES_YML_FILE_NAME = 'profiles.yaml'

DEFAULT_PROFILE_NAME = 'default'


def get_metadata_root_path(workspace_path):
    """Construct the path to a root metadata directory.

    :param workspace_path: The exact path to the root of a ckx_tools workspace
    :type workspace_path: str

    :returns: The path to the metadata root directory or None if workspace_path isn't a string
    :rtype: str or None
    """

    # TODO: Should calling this without a string just be a fatal error?
    if workspace_path is None:
        return None

    return os.path.join(workspace_path, METADATA_DIR_NAME)


def get_profiles_path(workspace_path):
    """Construct the path to a root metadata directory.

    :param workspace_path: The exact path to the root of a ckx_tools workspace
    :type workspace_path: str

    :returns: The path to the profile metadata directory or None if workspace_path isn't a string
    :rtype: str or None
    """

    if workspace_path is None:
        return None

    return os.path.join(workspace_path, METADATA_DIR_NAME, 'profiles')


def get_paths(workspace_path, profile_name, verb=None):
    """Construct the path to a metadata directory and verb-specific metadata file.

    Note: these paths are not guaranteed to exist. This function simply serves
    to standardize where these files should be located.

    :param workspace_path: The path to the root of a catkin workspace
    :type workspace_path: str
    :param profile_name: The ckx_tools metadata profile name
    :type profile_name: str
    :param verb: (optional) The ckx_tools verb with which this information is associated.

    :returns: A tuple of the metadata directory and the verb-specific file path, if given
    """

    # Get the root of the metadata directory
    profiles_path = get_profiles_path(workspace_path)

    # Get the active profile directory
    metadata_path = os.path.join(profiles_path, profile_name) if profile_name else None

    # Get the metadata for this verb
    metadata_file_path = os.path.join(metadata_path, '%s.yaml' % verb) if profile_name and verb else None

    return (metadata_path, metadata_file_path)


def find_enclosing_workspace(search_start_path):
    """Find a catkin workspace based on the existence of a ckx_tools
    metadata directory starting in the path given by search_path and traversing
    each parent directory until either finding such a directory or getting to
    the root of the filesystem.

    :search_start_path: Directory which either is a catkin workspace or is
    contained in a catkin workspace

    :returns: Path to the workspace if found, `None` if not found.
    """
    while search_start_path:
        # Check if marker file exists
        candidate_path = os.path.join(search_start_path, METADATA_DIR_NAME)
        if os.path.exists(candidate_path) and os.path.isdir(candidate_path):
            return search_start_path

        # Update search path or end
        (search_start_path, child_path) = os.path.split(search_start_path)
        if len(child_path) == 0:
            break

    return None

def find_enclosing_profile(search_start_path, workspace_hint):
    """Discover a parallel build folder backed by a profile if there
    is one to be discovered at or below the specified search start path
    hint. This is useful so that simply switching to the parallel build
    folder is enough to activate the various commands instead of having
    to switch to the active profile.

    :param str search_start_path: directory to start searching from
    :param str workspace_hint: should be from this workspace (may be None)
    :returns: name of the associated profile or `None` if not found
    """
    enclosing_workspace = find_enclosing_workspace(search_start_path)
    expected_workspace = find_enclosing_workspace(workspace_hint or search_start_path)
    if enclosing_workspace and \
       enclosing_workspace == expected_workspace and \
       enclosing_workspace != search_start_path:
        candidate_profile_name = os.path.basename(search_start_path)
        parent_path = os.path.abspath(os.path.join(search_start_path, os.pardir))
        while parent_path != enclosing_workspace:
            candidate_profile_name = os.path.basename(parent_path)
            parent_path = os.path.abspath(os.path.join(parent_path, os.pardir))
        if candidate_profile_name in get_profile_names(enclosing_workspace):
            return candidate_profile_name
    return None


def migrate_metadata(workspace_path):
    """Migrate metadata if it's out of date."""
    metadata_root_path = get_metadata_root_path(workspace_path)

    # Nothing there to migrate
    if not metadata_root_path or not os.path.exists(metadata_root_path):
        return

    # Check metadata version
    last_version = None
    current_version = pkg_resources.require("ckx_tools")[0].version
    version_file_path = os.path.join(metadata_root_path, 'VERSION')

    # Read the VERSION file
    if os.path.exists(version_file_path):
        with open(version_file_path, 'r') as metadata_version:
            last_version = metadata_version.read() or None
    if last_version != current_version:
        # Write the VERSION file
        with open(version_file_path, 'w') as metadata_version:
            metadata_version.write(current_version)
        migrate_metadata_version(workspace_path, last_version)


def migrate_metadata_version(workspace_path, old_version):
    """Migrate between metadata versions.
    """

    # Versions were added in 0.4.0, and the previously released version was 0.3.1
    if old_version is None:
        old_version = '0.3.1'

    old_version = tuple((int(i) for i in old_version.split('.')))
    metadata_root_path = get_metadata_root_path(workspace_path)
    new_profiles_path = os.path.join(metadata_root_path, 'profiles')

    # Restructure profiles directory
    if old_version < (0, 4, 0):

        for profile_name in os.listdir(metadata_root_path):

            if profile_name == 'profiles':
                continue

            profile_path = os.path.join(metadata_root_path, profile_name)
            if not os.path.isdir(profile_path):
                continue

            new_path = os.path.join(new_profiles_path, profile_name)
            shutil.move(profile_path, new_path)

    # Update metadata
    for profile_name in get_profile_names(workspace_path):
        for verb in ['config', 'build']:

            # Get the current metadata
            metadata = get_metadata(workspace_path, profile_name, verb)

            # Update devel layout for 0.3.1 -> 0.4.0
            if old_version < (0, 4, 0):
                isolate_devel = metadata.get('isolate_devel')
                if isolate_devel is not None:
                    del metadata['isolate_devel']
                devel_layout = ('isolated' if isolate_devel else 'merged')
                metadata['devel_layout'] = devel_layout

            # Save the new metadata
            update_metadata(workspace_path, profile_name, verb, metadata, no_init=True, merge=False)


def init_metadata_root(workspace_path, reset=False):
    """Create or reset a ckx_tools metadata directory with no content in a given path.

    :param workspace_path: The exact path to the root of a catkin workspace
    :type workspace_path: str
    :param reset: If true, clear the metadata directory of all information
    :type reset: bool
    """

    # Make sure the directory
    if not os.path.exists(workspace_path):
        raise IOError(
            "Can't initialize Catkin workspace in path %s because it does "
            "not exist." % (workspace_path))

    # Check if the desired workspace is enclosed in another workspace
    marked_workspace = find_enclosing_workspace(workspace_path)

    if marked_workspace and marked_workspace != workspace_path:
        raise IOError(
            "Can't initialize Catkin workspace in path %s because it is "
            "already contained in another workspace: %s." %
            (workspace_path, marked_workspace))

    # Construct the full path to the metadata directory
    metadata_root_path = get_metadata_root_path(workspace_path)

    # Check if a metadata directory already exists
    if os.path.exists(metadata_root_path):
        # Reset the directory if requested
        if reset:
            print("Deleting existing metadata from ckx_tools metadata directory: %s" % (metadata_root_path))
            shutil.rmtree(metadata_root_path)
            os.mkdir(metadata_root_path)
    else:
        # Create a new .ckx_tools directory
        os.mkdir(metadata_root_path)

    # Write the README file describing the directory
    with open(os.path.join(metadata_root_path, 'README'), 'w') as metadata_readme:
        metadata_readme.write(METADATA_README_TEXT)

    # Migrate the metadata, if necessary
    migrate_metadata(workspace_path)

    # Add a catkin ignore file so we can store package.xml files for cleaned packages
    if not os.path.exists(os.path.join(metadata_root_path, 'CATKIN_IGNORE')):
        open(os.path.join(metadata_root_path, 'CATKIN_IGNORE'), 'a').close()

def init_profile(workspace_path, profile_name, reset=False):
    """Initialize a profile directory in a given workspace.

    :param workspace_path: The exact path to the root of a ckx_tools workspace
    :type workspace_path: str
    :param profile_name: The ckx_tools metadata profile name to initialize
    :type profile_name: str
    """

    init_metadata_root(workspace_path)

    (profile_path, _) = get_paths(workspace_path, profile_name)

    # Check if a profile directory already exists
    if os.path.exists(profile_path):
        # Reset the directory if requested
        if reset:
            print("Deleting existing profile from ckx_tools profile directory: %s" % (profile_path))
            shutil.rmtree(profile_path)
            os.mkdir(profile_path)
    else:
        # Create a new .ckx_tools directory
        common.mkdir_p(profile_path)


def get_profile_names(workspace_path):
    """Get a list of profile names available to a given workspace.

    :param workspace_path: The exact path to the root of a ckx_tools workspace
    :type workspace_path: str

    :returns: A list of the available profile names in the given workspace
    :rtype: list
    """

    migrate_metadata(workspace_path)

    profiles_path = get_profiles_path(workspace_path)

    if os.path.exists(profiles_path):
        directories = next(os.walk(profiles_path))[1]

        return directories

    return []


def remove_profile(workspace_path, profile_name):
    """Remove a profile by name.

    :param workspace_path: The exact path to the root of a ckx_tools workspace
    :type workspace_path: str
    :param profile_name: The ckx_tools metadata profile name to delete
    :type profile_name: str
    """

    migrate_metadata(workspace_path)

    (profile_path, _) = get_paths(workspace_path, profile_name)

    if os.path.exists(profile_path):
        shutil.rmtree(profile_path)  # this is the contents in the .ckx_tools dir
    if profile_name != DEFAULT_PROFILE_NAME:
        shutil.rmtree(os.path.join(workspace_path, profile_name))  # this is the parallel build folder
    else:
        # clean up config files
        for filename in ['config.cmake', 'eclipse', 'setup.bash', 'konsole', 'gnome-terminal']:
            os.remove(os.path.join(workspace_path, filename))
        # TODO : should actually pull the default context here and grab the directories
        for d in ['build', 'devel', 'logs', 'install', 'docs']:
            shutil.rmtree(path=os.path.join(workspace_path, d), ignore_errors=True)


def set_active_profile(workspace_path, profile_name):
    """Set a profile in a given workspace to be active.

    :param workspace_path: The exact path to the root of a ckx_tools workspace
    :type workspace_path: str
    :param profile_name: The ckx_tools metadata profile name to activate
    :type profile_name: str
    """

    profiles_data = get_profiles_data(workspace_path)
    profiles_data['active'] = profile_name

    profiles_path = get_profiles_path(workspace_path)
    profiles_yaml_file_path = os.path.join(profiles_path, PROFILES_YML_FILE_NAME)
    with open(profiles_yaml_file_path, 'w') as profiles_file:
        yaml.dump(profiles_data, profiles_file, default_flow_style=False)


def get_active_profile(workspace_path):
    """Get the active profile name from a workspace path.

    :param workspace_path: The exact path to the root of a ckx_tools workspace
    :type workspace_path: str

    :returns: The active profile name
    :rtype: str
    """

    profiles_data = get_profiles_data(workspace_path)
    if 'active' in profiles_data:
        return profiles_data['active']

    return DEFAULT_PROFILE_NAME


def get_profiles_data(workspace_path):
    """Get the contents of the profiles file.

    This file contains information such as the currently active profile.

    :param workspace_path: The exact path to the root of a ckx_tools workspace
    :type workspace_path: str

    :returns: The contents of the root profiles file if it exists
    :rtype: dict
    """

    migrate_metadata(workspace_path)

    if workspace_path is not None:
        profiles_path = get_profiles_path(workspace_path)
        profiles_yaml_file_path = os.path.join(profiles_path, PROFILES_YML_FILE_NAME)
        if os.path.exists(profiles_yaml_file_path):
            with open(profiles_yaml_file_path, 'r') as profiles_file:
                return yaml.load(profiles_file)

    return {}


def get_metadata(workspace_path, profile, verb):
    """Get a python structure representing the metadata for a given verb.

    :param workspace_path: The exact path to the root of a catkin workspace
    :type workspace_path: str
    :param profile: The ckx_tools metadata profile name
    :type profile: str
    :param verb: The ckx_tools verb with which this information is associated
    :type verb: str

    :returns: A python structure representing the YAML file contents (empty
    dict if the file does not exist)
    :rtype: dict
    """

    migrate_metadata(workspace_path)

    (metadata_path, metadata_file_path) = get_paths(workspace_path, profile, verb)

    if not os.path.exists(metadata_file_path):
        return dict()

    with open(metadata_file_path, 'r') as metadata_file:
        result = yaml.load(metadata_file)
        return result


def update_metadata(workspace_path, profile, verb, new_data={}, no_init=False, merge=True):
    """Update the ckx_tools verb metadata for a given profile.

    :param workspace_path: The path to the root of a catkin workspace
    :type workspace_path: str
    :param profile: The ckx_tools metadata profile name
    :type profile: str
    :param verb: The ckx_tools verb with which this information is associated
    :type verb: str
    :param new_data: A python dictionary or array to write to the metadata file
    :type new_data: dict
    """

    migrate_metadata(workspace_path)

    (metadata_path, metadata_file_path) = get_paths(workspace_path, profile, verb)

    # Make sure the metadata directory exists
    if not no_init:
        init_metadata_root(workspace_path)
        init_profile(workspace_path, profile)

    # Get the curent metadata for this verb
    if merge:
        data = get_metadata(workspace_path, profile, verb)
    else:
        data = dict()

    # Update the metadata for this verb
    data.update(new_data)
    with open(metadata_file_path, 'w') as metadata_file:
        yaml.dump(data, metadata_file, default_flow_style=False)

    return data


def get_active_metadata(workspace_path, verb):
    """Get a python structure representing the metadata for a given verb.
    :param workspace_path: The exact path to the root of a catkin workspace
    :type workspace_path: str
    :param verb: The ckx_tools verb with which this information is associated
    :type verb: str

    :returns: A python structure representing the YAML file contents (empty
    dict if the file does not exist)
    :rtype: dict
    """

    active_profile = get_active_profile(workspace_path)
    get_metadata(workspace_path, active_profile, verb)


def update_active_metadata(workspace_path, verb, new_data={}):
    """Update the ckx_tools verb metadata for the active profile.

    :param workspace_path: The path to the root of a catkin workspace
    :type workspace_path: str
    :param verb: The ckx_tools verb with which this information is associated
    :type verb: str
    :param new_data: A python dictionary or array to write to the metadata file
    :type new_data: dict
    """

    active_profile = get_active_profile(workspace_path)
    update_active_metadata(workspace_path, active_profile, verb, new_data)
